/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2019  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QFont>
#include <QFontMetrics>
#include <QHBoxLayout>
#include <QLabel>
#include <QToolButton>
#include <QSize>
#include <QSizePolicy>
#include <QSlider>
#include <QVBoxLayout>
#include <QWidget>
#include <QScrollArea>
#include <QLineEdit>
#include <QSpinBox>
#include <QFrame>
#include <QIcon>
#include <QDir>
#include <QSettings>
#include <QCloseEvent>
#include <QTcpSocket>
#include <QKeyEvent>
#include <cmath>
#include <assert.h>

#include "../../pcdmonitor/share/displaywidget.h"
#include "../../pcdmonitor/share/ipwidget.h"
#include "pcdmonitorremote.h"

TPCDMonitorRemote::TPCDMonitorRemote(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QMainWindow(const_cast<QWidget*>(parent),flags)
{
	QFont								display_font;
	QFont								status_font;
	QFont								copyright_font;
	QLabel								*fontsize_label;
	QLabel								*copyright_label;
	QVBoxLayout							*widget_vlayout;
	QHBoxLayout							*widget_hlayout;
	QWidget								*central_widget;
	QWidget								*spacer_widget;
	QWidget								*spacer_widget_fixed_right;
	QWidget								*spacer_widget_fixed_left;
	QScrollArea							*scroll_area;
	QFrame								*separator_line;
	QSizePolicy							size_policy(QSizePolicy::MinimumExpanding,QSizePolicy::Preferred);
	QDir								home;
	
	home = QDir::home();
	
	if(!home.exists(".pcdmonitorremote"))
	{
		home.mkdir(".pcdmonitorremote");
	}
	
	home.cd(".pcdmonitorremote");
	
	d_settings = new QSettings(home.absolutePath() + "/Settings.ini",QSettings::IniFormat);

	size_policy.setHorizontalStretch(0);
	size_policy.setVerticalStretch(0);

	status_font.setWeight(QFont::DemiBold);

	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);
	
	display_font.setWeight(QFont::Medium);
	display_font.setFamily(QStringLiteral("Tahoma"));
	
	this->resize(1024,600);

	central_widget = new QWidget(this);
	this->setCentralWidget(central_widget);
	
	widget_vlayout = new QVBoxLayout(central_widget);
	widget_vlayout->setContentsMargins(6, 6, 6, 6);
	
	d_toolbar = new QWidget(central_widget);

	widget_hlayout = new QHBoxLayout(d_toolbar);
	widget_hlayout->setContentsMargins(6, 6, 6, 6);

	spacer_widget_fixed_left = new QWidget(d_toolbar);
    spacer_widget_fixed_left->setMinimumSize(5,0);
	widget_hlayout->addWidget(spacer_widget_fixed_left);

	d_power_button = new QToolButton(d_toolbar);
	d_power_button->setMinimumSize(QSize(96, 48));
	d_power_button->setIconSize(QSize(96, 48));
	d_power_button->setCheckable(false);
	widget_hlayout->addWidget(d_power_button);

	spacer_widget = new QWidget(d_toolbar);
	spacer_widget->setSizePolicy(size_policy);
	widget_hlayout->addWidget(spacer_widget);
	
	d_ip_widget = new TIPWidget(d_toolbar);
	widget_hlayout->addWidget(d_ip_widget);

	fontsize_label = new QLabel(d_toolbar);
	widget_hlayout->addWidget(fontsize_label);

	d_font_slider = new QSlider(d_toolbar);
    d_font_slider->setMinimumSize(81,0);
    d_font_slider->setMinimum(6);
	d_font_slider->setMaximum(50);
	d_font_slider->setOrientation(Qt::Horizontal);
	widget_hlayout->addWidget(d_font_slider);

	spacer_widget_fixed_right = new QWidget(central_widget);
    spacer_widget_fixed_right->setMinimumSize(5,0);
	widget_hlayout->addWidget(spacer_widget_fixed_right);
	
	widget_vlayout->addWidget(d_toolbar);

	scroll_area = new QScrollArea(central_widget);
	scroll_area->setWidgetResizable(true);
 	widget_vlayout->addWidget(scroll_area);

	d_display_widget = new TDisplayWidget;	
	d_display_widget->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	d_display_widget->setWordWrap(false);
    scroll_area->setWidget(d_display_widget);
	
	d_status_text = new QLabel(central_widget);
	d_status_text->setFont(status_font);
	widget_vlayout->addWidget(d_status_text);
	
	separator_line = new QFrame(this);
	separator_line->setFrameShape(QFrame::HLine);
	separator_line->setFrameShadow(QFrame::Sunken);
	widget_vlayout->addWidget(separator_line);

	copyright_label = new QLabel(central_widget);
	copyright_label->setFont(copyright_font);
	widget_vlayout->addWidget(copyright_label);
	
	d_tcp_socket = new QTcpSocket(this);
	
	d_ip_widget->Set_IP_Address(d_settings->value("IP_Address","127.0.0.1").toString());
	d_ip_widget->Set_IP_Port(d_settings->value("IP_Port",2112).toInt());
	d_font_slider->setValue(d_settings->value("Font_Size",18).toInt());

	// defaults
	display_font.setPointSize(18);
	d_display_widget->setFont(display_font);
	
	this->Update_Font(d_font_slider->value());
	
	d_power_button->setIcon(QIcon(QStringLiteral(":/images/button_off.png")));
	d_connection_status = TPCDMonitorRemote::NO_CONNECTION;
	this->Set_Status_Text("INFO:  Disconnected.");

	this->setWindowTitle(QStringLiteral("PC-DMIS Monitor Remote - Version 2.1"));
	
	fontsize_label->setText(QStringLiteral("Font Size:"));

	copyright_label->setText(QStringLiteral("Copyright (C) 2019 Select Calibration Incorporated.  www.selectcalibration.ca"));
	

	connect(d_font_slider,SIGNAL(valueChanged(int)),this,SLOT(Update_Font(int)));
	connect(d_power_button,SIGNAL(clicked(void)),this,SLOT(Connect_Disconnect(void)));
	
	connect(d_tcp_socket, SIGNAL(readyRead(void)),this,SLOT(Data_Ready(void)));
	connect(d_tcp_socket, SIGNAL(error(QAbstractSocket::SocketError)),this,SLOT(Data_Error(QAbstractSocket::SocketError)));
	connect(d_tcp_socket, SIGNAL(connected(void)),this,SLOT(Connected(void)));
}

TPCDMonitorRemote::~TPCDMonitorRemote(void)
{
}

void TPCDMonitorRemote::Update_Font(
	int									point_size)
{
	QFont								font(d_display_widget->font());
	
	font.setPointSize(point_size);
	d_display_widget->setFont(font);
}

void TPCDMonitorRemote::Connect_Disconnect(void)
{
	if(d_connection_status == TPCDMonitorRemote::NO_CONNECTION)
	{
		d_power_button->setIcon(QIcon(QStringLiteral(":/images/button_warn.png")));

		d_connection_status = TPCDMonitorRemote::CONNECTING;
		this->Set_Status_Text("INFO:  Connecting...");
		
		d_tcp_socket->abort();
		d_tcp_socket->connectToHost(d_ip_widget->IP_Address(),d_ip_widget->IP_Port());
		
		d_ip_widget->Set_Enabled(false);
	}
	else
	{
		d_connection_status = TPCDMonitorRemote::NO_CONNECTION;
		d_power_button->setIcon(QIcon(QStringLiteral(":/images/button_off.png")));
		
		this->Set_Status_Text("INFO:  Disconnected.");
		d_tcp_socket->abort();
		
		d_ip_widget->Set_Enabled(true);
	}
}

void TPCDMonitorRemote::Connected(void)
{
	d_power_button->setIcon(QIcon(QStringLiteral(":/images/button_on.png")));
	
	d_connection_status = TPCDMonitorRemote::CONNECTED;
	this->Set_Status_Text("INFO:  Connected.");
	
	d_receiving_status = false;
}

void TPCDMonitorRemote::Data_Ready(void)
{
	QString								data;
	QString								status_data;
	int									index;
	int									status_index_start;
	int									status_index_end;
	
	data = QString(d_tcp_socket->readAll());
	
	status_index_start = data.lastIndexOf(char(0x01));
	status_index_end = data.lastIndexOf(char(0x02));
	
	if(!(status_index_start < 0) && !(status_index_end < 0))
	{
		// data contains entire status message.
		assert(status_index_end > status_index_start);
		
		status_data = data.mid(status_index_start+1);
		status_data.truncate(status_index_end - status_index_start - 1);
		
		this->Set_Status_Text(status_data);
		
		data.remove(status_index_start,status_index_end - status_index_start + 1);
		d_receiving_status = false;
	}
	else if(!(status_index_start < 0))
	{
		// start but no end
		status_data = data.mid(status_index_start+1);
		this->Set_Status_Text(status_data);
		data.truncate(status_index_start);
		d_receiving_status = true;
	}
	else if(!(status_index_end < 0))
	{
		if(d_receiving_status)
		{
			status_data = data.mid(0,status_index_end);
			this->Add_Status_Text(status_data);
			
			data.truncate(status_index_end+1);
		}
		
		d_receiving_status = false;
	}
	
	index = data.lastIndexOf('\f');
	
	if(!(index < 0))
	{
		// ignore first part of data.
		data = data.mid(index+1);
		
		this->Set_Display_Text(data);
	}
	else
	{
		this->Add_Display_Text(data);
	}
}

void TPCDMonitorRemote::Data_Error(
	QAbstractSocket::SocketError		socket_error)
{
	d_power_button->setIcon(QIcon(QStringLiteral(":/images/button_err.png")));
	d_connection_status = TPCDMonitorRemote::NO_CONNECTION;
	
	switch (socket_error)
	{
		case QAbstractSocket::RemoteHostClosedError:
			this->Set_Status_Text("ERR:  Connection closed remotely.");
			break;
			
		case QAbstractSocket::HostNotFoundError:
			this->Set_Status_Text("ERR:  Host not found.");
			break;
			
		case QAbstractSocket::SocketTimeoutError:
			this->Set_Status_Text("ERR:  Connection timeout.");
			break;
			
		case QAbstractSocket::ConnectionRefusedError:
			this->Set_Status_Text("ERR:  Connection refused.");
			break;
			
		default:
			this->Set_Status_Text(QString("ERR:  Socket error number: %1.").arg(static_cast<int>(socket_error)));
			break;
	}

    d_ip_widget->Set_Enabled(true);
}

void TPCDMonitorRemote::closeEvent(
	QCloseEvent 						*event)
{
	if(d_connection_status != TPCDMonitorRemote::NO_CONNECTION)
	{
		this->Connect_Disconnect();
	}
	
	d_settings->setValue("IP_Address", d_ip_widget->IP_Address());
	d_settings->setValue("IP_Port", d_ip_widget->IP_Port());
	d_settings->setValue("Font_Size", d_font_slider->value());

	event->accept();
}

void TPCDMonitorRemote::keyReleaseEvent(
	QKeyEvent							*event)
{
	QMainWindow::keyPressEvent(event);
	
#ifdef Q_OS_MAC
	if(event->key() == Qt::Key_Control) // command key
#else
	if(event->key() == Qt::Key_Menu)
#endif
	{
		if(d_toolbar->isVisible())
		{
			d_toolbar->setVisible(false);
		}
		else
		{
			d_toolbar->setVisible(true);
		}
		
		event->accept();
	}
}

void TPCDMonitorRemote::Set_Display_Text(
	const QString						&display_text)
{
	d_display_widget->setText(display_text);
}

void TPCDMonitorRemote::Set_Status_Text(
	const QString						&status_text)
{
	d_status_text->setText(status_text);
}

void TPCDMonitorRemote::Add_Display_Text(
	const QString						&text)
{
	QString								widget_text;
	
	widget_text = d_display_widget->text();
	
	if(widget_text.length())
	{
		widget_text.push_back('\n');
	}
	widget_text += text;
	
	d_display_widget->setText(widget_text);
}

void TPCDMonitorRemote::Add_Status_Text(
	const QString						&text)
{
	QString								widget_text;
	
	widget_text = d_status_text->text() + text;
	
	d_status_text->setText(widget_text);
}






